@@ -88,15 +88,16 @@ impl FreeList {
         let mut freelist = [0; 32];
         for p in &mut freelist.iter_mut().enumerate() {
             let pos = 8 * p.0;
-            (*p.1) = read_u64(&buffer[pos..pos + 8]);
+            (*p.1) = read_u64(&buffer[pos..(pos + 8)]);
         }
         freelist
     }
 
     pub fn write_free_list(list: &[u64]) -> [u8; 259] {
         let mut buffer = [0; 259];
-        for (pos, page) in list.iter().enumerate() {
-            write_u64(&mut buffer[pos..pos + 8], *page);
+        for (index, page) in list.iter().enumerate() {
+            let pos = index * 8;
+            write_u64(&mut buffer[pos..(pos + 8)], *page);
         }
         buffer
     }
@@ -138,7 +139,18 @@ pub struct Allocator {
 impl Allocator {
     pub fn new(dr: Box<dyn Device>, config: &Config, page: u64) -> PRes<Allocator> {
         let mut pg = dr.load_page(page)?;
-        let freelist = FreeList::read(&mut pg)?;
+        let mut freelist = FreeList::read(&mut pg)?;
+
+        // This may cause disk leaks in case of crash, but
+        // keep the system safe
+        for page in freelist.list {
+            if page != 0 {
+                let lp = dr.load_page(page)?;
+                if !lp.is_free()? {
+                    freelist.list[lp.get_size_exp() as usize] = 0;
+                }
+            }
+        }
 
         let cache_size = config.cache_size();
         Ok(Allocator {
@@ -160,11 +172,26 @@ impl Allocator {
     }
 
     pub fn load_page_not_free(&self, page: u64) -> PRes<Option<ReadPage>> {
-        let load = self.read_page_int(page)?;
-        if load.is_free()? {
-            Ok(None)
+        {
+            let mut cache = self.cache.lock()?;
+            if let Some(pg) = cache.get(page) {
+                if pg.is_free()? {
+                    return Ok(None);
+                } else {
+                    return Ok(Some(pg));
+                }
+            }
+        }
+        if let Some(load) = self.disc.load_page_if_exists(page)? {
+            if load.is_free()? {
+                Ok(None)
+            } else {
+                let mut cache = self.cache.lock()?;
+                cache.put(page, load.clone_read());
+                Ok(Some(load))
+            }
         } else {
-            Ok(Some(load))
+            Ok(None)
         }
     }
 
@@ -385,7 +412,7 @@ impl UpdateList for FreeList {
 
 #[cfg(test)]
 mod tests {
-    use super::{Allocator, Cache};
+    use super::{Allocator, Cache, FreeList};
     use crate::{
         config::Config,
         discref::{DiscRef, PageOps, ReadPage},
@@ -511,4 +538,16 @@ mod tests {
             None => assert!(false),
         }
     }
+
+    #[test]
+    fn test_read_write_free_list() {
+        let mut list = [0u64; 32];
+        for p in 0..list.len() {
+            // Random value do not matter
+            list[p] = p as u64 + 10 * 3;
+        }
+        let buf = FreeList::write_free_list(&list);
+        let read_list = FreeList::read_free_list(&buf);
+        assert_eq!(list, read_list);
+    }
 }
